# encoding: UTF-8

import os.path
import random
import shutil
import traceback
import test


def perform_sanity_checks(module, top_levels_whitelist, occurrence_whitelist, raise_errors=True):
    # This will get executed for getting the object names, but
    # in such cases the sanity check must not be performed to
    # avoid that its actions affect the object name related
    # activity, like copying an object name, or so. We can
    # detect this because in these cases the traceback's first
    # frame does not end with test.py:
    if not traceback.extract_stack()[0][0].endswith("test.py"):
        return True

    # We need to work with a copy of this module, to avoid
    # that recording adds duplicate object names.
    #
    # But only perform our sanity checks in case of the
    # original file, not the copy that we create and load next:
    if os.path.basename(module.__file__) == "tmp_om_sanity_check.py":
        return True

    return __perform_sanity_checks(module, top_levels_whitelist, occurrence_whitelist, raise_errors)


def __perform_sanity_checks(module, top_levels_whitelist, occurrence_whitelist, raise_errors=True):
    fn = module.__file__

    # We may be getting the binary .pyc file reported here,
    # use the .py file instead then:
    if fn.endswith(".pyc"):
        fn = fn[:-1]

    fn2 = "%s/tmp_om_sanity_check.py" % (os.path.dirname(fn))
    shutil.copyfile(fn, fn2)
    import tmp_om_sanity_check

    try:
        test.fixateResultContext(2)
        result = sanity_check_occurrence(
            tmp_om_sanity_check,
            module.__name__,
            occurrence_whitelist)
    
        result = sanity_check_top_levels(
            tmp_om_sanity_check,
            module.__name__,
            top_levels_whitelist) and result
    
        if raise_errors:
            raise NameError("ERROR: Object map sanity checks failed. See details above. (Module: %s" % module.__name__)
        
        return result
    finally:
        test.restoreResultContext()


def sanity_check_occurrence(object_map_module, original_module_name, whitelist):
    unexpected_names_with_occurrence = []
    for n in get_object_names(object_map_module):
        v = get_real_name(object_map_module, n)
        if "occurrence" in v and n not in whitelist:
            unexpected_names_with_occurrence.append(n)

    if len(unexpected_names_with_occurrence) == 0:
        return True

    test.startSection("ERROR: %s unexpected object name(s) with \"occurrence\" found:" % len(unexpected_names_with_occurrence))
    try:
        log()
        for n in unexpected_names_with_occurrence:
            log("    %s" % n, repr(get_real_name(object_map_module, n)))
        log()
        log()
        log("%s whitelisted names with \"occurrence\":" % len(whitelist))
        log()
        if len(whitelist) > 0:
            for n in whitelist:
                if hasattr(object_map_module, n):
                    log("    %s" % n, repr(get_real_name(object_map_module, n)))
                else:
                    log("    %s" % n, "Does not exist in %s" % object_map_module)
        else:
            log("    (none)")
        log()
        log()
        log("Options at this point:")
        log()
        log("    1) Inspect the objects identified by the mentioned object names, and")
        log("       check if the objects can be identified without resorting to \"occurrence\".")
        log()
        log("    2) Whitelist the unexpected names with \"occurrence\".")
        log()
        log("    3) Contact QA development support.")
        log()
        log()
    finally:
        test.endSection()

    return False



def sanity_check_top_levels(object_map_module, original_module_name, whitelist):
    unexpected_top_levels = []
    for n in get_top_level_object_names(object_map_module):
        if n not in whitelist:
            unexpected_top_levels.append(n)

    if len(unexpected_top_levels) == 0:
        return True

    test.startSection("ERROR: %s unexpected top level object name(s) found:" % len(unexpected_top_levels))
    try:
        log()
        for n in unexpected_top_levels:
            log("    %s" % n, repr(get_real_name(object_map_module, n)))
        log()
        log()
        log("%s whitelisted top level object names:" % len(whitelist))
        log()
        if len(whitelist) > 0:
            for n in whitelist:
                if hasattr(object_map_module, n):
                    log("    %s" % n, repr(get_real_name(object_map_module, n)))
                else:
                    log("    %s" % n, "Does not exist in %s" % object_map_module)
        else:
            log("    (none)")
        log()
        log()
        log("Options at this point:")
        log()
        log("    1) Consolidate the unexpected top level object names which identify")
        log("       the same object. Use wildcards or regular expressions where suitable.")
        log()
        log("    2) Whitelist the unexpected top level object names.")
        log()
        log("    3) Contact QA development support.")
        log()
        log()
    finally:
        test.endSection()

    return False


def log(msg="", details=None):
    if details is None:
        test.warning(msg)
    else:
        test.warning(msg, details)


def get_real_name(module, symbolic_name):
    return getattr(module, symbolic_name)


def get_top_level_object_names(module):
    top_levels = []
    names_ = get_object_names(module)
    for n in names_:
        v = get_real_name(module, n)
        if is_top_level_real_name(v):
            top_levels.append(n)
    return top_levels


def get_object_names(module):
    res = []
    for n in dir(module):
        if n == "__builtins__":
            continue
        v = get_real_name(module, n)
        if type(v) == dict:
            res.append(n)
    return res


def is_top_level_real_name(real_name):
    if type(real_name) != dict:
        return False

    # This is a simplistic definition of what makes an
    # real name a name for a top level object, but it
    # will do for most object names generated by Squish:
    return "container" not in real_name \
        and "window" not in real_name \
        and "buddy" not in real_name \
        and "parentItem" not in real_name
